import pylab
from basic import *
from numpy import *

def initial_sin(p):
    """Sin(\mu) with \mu \in [-pi/2;pi/2], such that one half is at each side respectively."""
    p.import_boundary(sin(0.5*pi*p.array_mu));
    p.initial_method="sin";

def initial_sin_sign(p):
    """sin(pi*x) for \mu>0."""
    p.import_boundary(sin(pi*p.array_mu)*(1+sign(p.array_mu))/2.)
    p.initial_method="half sin";

def initial_rect(p):
    """Constant 1, i.e. isotropic, inflow boundaries."""
    p.import_boundary(ones(p.n_mu))
    p.initial_method="rect";

def initial_sign(p):
    """One directional isotropic inflow boundary conditions."""
    p.import_boundary((1+sign(p.array_mu))/2.);
    p.initial_method="sign";

def initial_beam(p):
    """Roughly translates to a straight forward (mu~=1) electron beam."""
    b = p.array_mu.copy();
    for i in range(p.n_mu):
        if b[i] >= 0.995:
            b[i]=1.;
        else:
            b[i]=0.;
    b[-1]=1.;

    p.import_boundary(b)
    p.initial_method="forward beam";

def initial_beam2(p):
    """A broader simple beam aimed at about mu=.5."""
    b = p.array_mu.copy();
    for i in range(p.n_mu):
        if b[i] > 0.25 and b[i] <0.76:
            b[i]=1.;
        else:
            b[i]=0.;

    p.import_boundary(b)
    p.initial_method="broad beam";

def initial_beam3(p):
    """Narrow Gaussian distributed beam aimed with mu=.5."""
    b = p.array_mu.copy();
    for i in range(p.n_mu):
        if p.array_mu[i]<0:
            b[i]=0.;
            continue
        b[i]=exp(-(b[i]-0.5)**2/1e-2)
    p.import_boundary(b)
    p.initial_method="gaussian beam";


def initial_gauss(p):
    """Simply Gaussian distribution over all of mu."""
    b = exp(-1000*(p.array_mu-0.5)**2)*(1+sign(p.array_mu))/2.;
    p.import_boundary(b)
    p.initial_method="gauss";
