#!/usr/bin/env python3

#custom python modules
from basic import *
from data import *
from solvers import *
from boundaries import *

## general available python modules
import sys
import os
import signal
import gc
gc.enable()
from multiprocessing import Pool,cpu_count,Value
from time import sleep
from random import randint
## non default python packages
try:
    import setproctitle
except ImportError:
    def set_procname(title):
        return;
else:
    set_procname=setproctitle.setproctitle;

try:
    from psutil import cpu_percent
except ImportError:
    def cpu_percent(interval=1):
        return;

global title,counter,number,pool
title="comparison"
pool=None

def local_shutdown(signal=None,stack_frame=None):
    """Terminate all child processes in the multiprocessing pool."""
    global pool
    print('received signal',signal)
    if pool != None:
        pool.terminate()
        pool.join()
signal.signal(signal.SIGTERM, local_shutdown)

def job_spreading():
    """Due to multithreading the load of the entire system can become quite high. Therefor let us save system resources by waiting for a time window."""
    #sleep(randint(0,100)/10.) ## Spread the start of the given functions a little bit. This is only disadvantageous fr small grids.
    return
    ## This section is not necessary, as MKL itself has the env variabel MKL_DYNAMIC=true, which should do something similar.
    #load=cpu_percent(randint(0,100)/10.)
    #cores=cpu_count()
    #while load > 100*(1-1/cores): #and load[1] > 1.25*cores: ## Add ram check
        #set_procname('Waiting.')
        #sleep(5)

def feedback(fdbck="",increment=1):
    """In conjunction with the global variables title, counter and number the global progress is measured and printed to stdout. The same information is also indicated in the process name, such that that information may be found easily in any provess overview"""
    global counter,number,title
    counter.value+=increment;
    local_name=title+" - {0}/{1}={2:3.1f}%".format(counter.value,number,100*counter.value/number)
    if isinstance(fdbck,str) and fdbck!="": local_name+=" - "+fdbck
    set_procname(local_name)
    #sys.stdout.write("Processing {0}/{1} - {2:3.1f}%\r".format(counter.value,number,100*counter.value/number),)
    sys.stdout.write(local_name.ljust(30)+"\r")

def one_comparison(xi=None,boundary=None,n=None,solvrs=None,param=parameters,compute_reference=False,inv_reference=True,output=None):
    """This function computes solutions with all the solvers for one given set of parameters."""
    global counter,number,title
    storage = data();
    if os.path.exists(output+'.json'):
        storage.import_json(output+'.json')
    job_spreading()
    # initialize the grid data, we only use square grids as of now.
    p=param(n,n,xi=xi)
    boundary(p);

    ## This check is done apriori to identify the leftover methods. The reference results should no be used for comparison at the moment, because they are saved as a list of string number with less precission. The saved data can easily be used for plotting.
    solvrs_tmp=[];
    for solv in solvrs:
        if storage.get_state(xi=xi,boundary=boundary,n_mu=n,n_z=n,solver=solv,parameter=param,kappa=None,lmbda=None) != 0:
            feedback('Skipping')
            continue
        solvrs_tmp.append(solv)
    solvrs=solvrs_tmp;

    # An accurate solution as a reference.
    #import pdb; pdb.set_trace()
    missing_reference = storage.get_state(xi=xi,boundary=boundary,n_mu=n,n_z=n,parameter=param,flags=['reference'],kappa=None,lmbda=None) == 0
    compute_reference = compute_reference and (solvrs == [] or missing_reference)
    if compute_reference: #storage.get_state(xi=xi,boundary=boundary,n_mu=n,n_z=n,parameter=param,flags=['reference'],kappa=None,lmbda=None) == 0:
        if n<128 and inv_reference: # the exact solver is for most circumstances to slow.
            reference=solver(p)
        else:
            reference=GMRES(p,tol=1e-8)
        feedback('REF-'+reference.get_feedback())
        reference.solve()
        if missing_reference: storage.append_solver(reference,save_grid=True)
        if isinstance(output,str): storage.export_json(path=output)
        feedback('REF-'+reference.get_feedback(),increment=0)
    else:
        feedback('Skipping')

    # Loop throught the solvers, which we want to investigate.
    for solv in solvrs:
        computer=solv(p)
        feedback(computer.get_feedback())
        computer.solve()
        if compute_reference and reference.state == 1: computer.compare(reference)
        storage.append_solver(computer,save_grid=False)
        if isinstance(output,str): storage.export_json(path=output)
        feedback(computer.get_feedback(),increment=0)

        del computer;
        gc.collect();

    del p
    if compute_reference: del reference
    gc.collect()

    set_procname(title+' - limbo')
    if isinstance(output,str):
        del storage
        gc.collect()
    else:
        return True #storage

def comparison(dict):
    """Just a wrapper around one_comparison in order to use it within a multiprocesses map function."""
    return one_comparison(**dict)


def do_comparison(json_out="simulation_data", ## The output base name.
                  params=[parameters], ## The possible parameter classes represent different stencils and are explained in more detail in basic.py
                  boundaries=[initial_rect], ## The dirichlet inflow boundary conditions to use. They should be defined in boundaries.py.
                  xi_list=[1.0], ## A list of all xi parameters to compute with.
                  grid=[32],#list(range(352,31,-32)),#[32,64,96,128,160,192,224,256,288,320,352], ## A list of grid sizes to the used square grids.
                  solvrs=[GMRES], ## A list of all the solvers to use besides the reference solution.
                  cores=-1 ## The number of subprocesses to used. -1 defaults to the number of available cores.
                  ):
    ## This function generates all the parameter combinations and thereafter starts all the specific computations. As default these processes are run in parallel.
    global counter,number,pool

    ## Loop through many grid sizes with the given boundary configurations.
    if cores==-1: cores=cpu_count()
    storage = data()
    arg_list=[]
    store=[]

    ## Generate one list with all possible parameter combinations.
    if not os.path.exists(json_out): os.makedirs(json_out)
    for xi in xi_list:
        for n in grid: ##Start with the slowest to compute.
            for param in params:
                for boundary in boundaries: # refer to boundaries.py for boundaries.
                    path=json_out+"/xi_{0}-n_{1}-{2}-{3}".format(xi,n,boundary.__name__,param.__name__)
                    arg_list.append({"xi": xi, "boundary": boundary, "n": n, "solvrs": solvrs, "param": param, "output": path})
    #import pdb; pdb.set_trace()
    ## Set the process counter und total number in order to meassure the progress.
    counter=Value('i',0)
    number=len(arg_list)*(len(solvrs)+1)

    ## The processes are started here.
    #pool=Pool(cores)
    #store=pool.imap_unordered(comparison,arg_list)
    #pool.close();
    #pool.join()
    #pool.terminate()
    for argment in arg_list: store.append(comparison(argment)) ## the simple single thread work queue - makes debuggung easier.

    ## The evaluation of all stored results. Every worker process exports it's own unique file.
    #for shop in store: storage.import_data(shop);
    #storage.export_json(path=json_out)
    return

if __name__ == "__main__":
    """The main function which is used, when this script is executed."""
    set_procname(title)
    out_path = "simulation_data2"
    params=[parameters]#,parameters_high]
    boundaries=[initial_rect,initial_beam3]
    xi_list=[1.]#0.125,0.15,0.175,0.2,0.25,0.375,0.5,0.75,1.0,1.25,1.5,2.0,3.0,4.0,5.0,6.0]
    grid=list(range(32,257,32))#list(range(32,353,32))
    solvrs=[GMRES_julia]#GMRES_right_mu_prec,GMRES_right_prec_matrixfree,Richardson_right_prec_matrixfree]

    do_comparison(out_path,params,boundaries,xi_list,grid,solvrs,cores=-1);
